/* 
   cadaver, command-line DAV client
   Copyright (C) 1999-2000, Joe Orton <joe@orton.demon.co.uk>, 
   except where otherwise indicated.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   $Id: commands.c,v 1.57 2000/08/03 08:55:40 joe Exp $
*/

/* Some UI guidelines:
 *  1. Use dispatch, or out_* to do UI. This makes it CONSISTENT.
 *  2. Get some feedback on the screen before making any requests
 *     to the server. Tell them what is going on: remember, on a slow
 *     link or a loaded server, a request can take AGES to return.
 */

#include <config.h>

#include <sys/types.h>

#include <sys/stat.h>
#include <sys/wait.h>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#include <fcntl.h>
#include <errno.h>

/* readline requires FILE *, silly thing */
#include <stdio.h>

#ifdef HAVE_READLINE_H
#include <readline.h>
#elif HAVE_READLINE_READLINE_H
#include <readline/readline.h>
#endif

#ifdef NEED_SNPRINTF_H
#include "snprintf.h"
#endif

#include <string_utils.h>
#include <http_request.h>
#include <http_basic.h>
#include <dav_basic.h>
#include <dav_props.h>
#include <uri.h>
#include <dav_locks.h>
#include <ne_alloc.h>

#include "i18n.h"
#include "dates.h"
#include "basename.h"
#include "dirname.h"
#include "cadaver.h"
#include "commands.h"
#include "options.h"

int yesno(void);

/* Local variables */
int child_running; /* true when we have a child running */

/* Command alias mappings */
const static struct {
    enum command_id id;
    const char *name;
} command_names[] = {
    /* The direct mappings */
#define C(n) { cmd_##n, #n }
    C(ls), C(cd), C(quit), C(open), C(close), C(set), C(unset), C(pwd),
    C(help), C(put), C(get), C(mkcol), C(delete), C(move), C(copy),
    C(less), C(cat), C(lpwd), C(lcd), C(lls), C(echo), C(quit),
    C(mget), C(mput), C(rmcol), C(lock), C(unlock), C(discover), C(steal),
    C(chexec), C(showlocks), C(version), 
#if 0
C(propedit), 
#endif
C(propnames), C(edit),
#undef C
    /* And now the real aliases */
    { cmd_less, "more" }, { cmd_mkcol, "mkdir" }, 
    { cmd_delete, "rm" }, { cmd_copy, "cp"}, { cmd_move, "mv" }, 
    { cmd_help, "h" }, { cmd_help, "?" },
    { cmd_quit, "exit" }, { cmd_quit, "bye" },
    { cmd_unknown, NULL }
};    

/* Tell them we are doing 'VERB' to 'NOUN'.
 * (really 'METHOD' to 'RESOURCE' but hey.) */
void out_start(const char *verb, const char *noun)
{
    output(o_start, "%s `%s':", verb, noun);
}

void out_success(void)
{
    output(o_finish, _("succeeded.\n"));
}

void out_result(int ret)
{
    switch (ret) {
    case HTTP_OK:
	out_success();
	break;
    case HTTP_AUTH:
	output(o_finish, _("authentication failed.\n"));
	break;
	/* TODO: others */
    case HTTP_CONNECT:
	output(o_finish, _("could not connect to server.\n"));
	break;
    case HTTP_TIMEOUT:
	output(o_finish, _("connection timed out.\n"));
	break;
    default:
	output(o_finish, _("failed:\n%s\n"), http_get_error(session));
	break;
    }
}

int out_handle(int ret)
{
    out_result(ret);
    return (ret == HTTP_OK);
}

/* The actual commands */

void init_locking(const char *lockstore)
{
    DEBUG(DEBUG_LOCKS, "init locking.\n");
    lock_session = dav_lock_register(session);
    /* TODO: read in lock list from ~/.davlocks */
}

void finish_locking(const char *lockstore)
{
    DEBUG(DEBUG_LOCKS, "Finish locking.\n");
    /* TODO: write out lock list to ~/.davlocks */
}

/* Auxiliaries */

static void 
map_multi(void (*func)(const char *), int argc, const char *argv[]);

static void simple_copy(const char *from, const char *to);
static void simple_put(const char *local, const char *remote);
static void simple_move(const char *src, const char *dest);

static void do_copymove(int argc, const char *argv[],
			const char *v1, const char *v2,
			void (*cb)(const char *, const char *));

static char *clever_path(const char *path, const char *src, 
			 const char *dest);

static char *choose_pager(void);

static void dispatch(const char *verb, const char *filename, 
		     int (*func)(http_session *, const char *), 
		     const char *pass);

static void display_help_message(void);

static void print_lock(struct dav_lock *lock);

#ifdef HAVE_LIBREADLINE

/* Command name generator for readline.
 * Copied almost verbatim from the info doc */
char *command_generator(char *text, int state)
{
    static int index, len;
    const char *name;

    if (!state) {
	index = 0;
	len = strlen(text);
    }

    while ((name = command_names[index].name) != NULL) {
	index++;
	if (strncmp(name, text, len) == 0) {
	    return ne_strdup(name);
	}
    }
    return NULL;
}

#endif

const struct command *get_command(const char *name)
{
    int n, m;
    for(n = 0; command_names[n].name != NULL; n++) {
	if (strcasecmp(command_names[n].name, name) == 0) {
	    for(m = 0; commands[m].id != cmd_unknown; m++) {
		if (commands[m].id == command_names[n].id)
		    return &commands[m];
	    }
	    return NULL;
	}
    }
    return NULL;
}


/* A bit like Haskell's map function... applies func to each
 * of the first argc items in argv */
static void 
map_multi(void (*func)(const char *), int argc, const char *argv[])
{
    int n;
    for(n = 0; n < argc; n++) {
	(*func)(argv[n]);
    }
}

void multi_mkcol(int argc, const char *argv[])
{
    map_multi(execute_mkcol, argc, argv);
}

void multi_delete(int argc, const char *argv[])
{
    map_multi(execute_delete, argc, argv);
}

void multi_rmcol(int argc, const char *argv[])
{
    map_multi(execute_rmcol, argc, argv);
}

void multi_less(int argc, const char *argv[])
{
    map_multi(execute_less, argc, argv);
}

void multi_cat(int argc, const char *argv[])
{
    map_multi(execute_cat, argc, argv);
}

static void dispatch(const char *verb, const char *filename, 
		     int (*func)(http_session *, const char *), const char *arg)
{
    out_start(verb, filename);
    out_result((*func)(session, arg));
}

static char * getowner(void)
{
#if 0
    /* mod_dav breaks if you submit an owner */
    return ne_strdup("");
#else
    char *ret, *owner = get_option(opt_lockowner);
    if (owner) {
	CONCAT3(ret, "<href>", owner, "</href>");
	return ret;
    } else {
	return NULL;
    }
#endif
}

/* Resolve path, appending trailing slash if resource is a
 * collection. This function MAY make a request to the server.
 * TODO: caching, so it MAY not, too.
 */
static char *true_path(const char *res)
{
    char *full;
    full = resolve_path(path, res, false);
    if (is_collection(full)) {
	if (!uri_has_trailing_slash(full)) {
	    char *tmp;
	    CONCAT2(tmp, full, "/");
	    free(full);
	    full = tmp;
	}
    }
    return full;
}

static const char *get_lockscope(enum dav_lock_scope s)
{
    switch (s) {
    case dav_lockscope_exclusive: return _("exclusive");
    case dav_lockscope_shared: return _("shared");
    default: return _("unknown");
    }
}

static const char *get_locktype(enum dav_lock_type t)
{
    if (t == dav_locktype_write) {
	return _("write");
    } else {
	return _("unknown");
    }
}

static const char *get_timeout(int t)
{
    static char buf[128];
    switch (t) {
    case DAV_TIMEOUT_INFINITE: return _("infinite");
    case DAV_TIMEOUT_INVALID: return _("invalid");
    default:
	sprintf(buf, "%d", t);
	return buf;
    }
}

static char *get_depth(int d)
{
    switch (d) {
    case DAV_DEPTH_INFINITE:
	return _("infinity");
    case 0:
	/* TODO: errr... do I need to i18n'ize numeric strings??? */
	return "0";
    case 1:
	return "1";
    default:
	return _("invalid");
    }
}

static void print_lock(struct dav_lock *lock)
{
    printf(_("Lock <%s> on %s:\n"
	     "  Scope: %s  Type: %s  Timeout: %s\n"
	     "  Owner: %s  Depth: %s\n"), 
	     lock->token, lock->uri?lock->uri:_("(unknown)"),
	     get_lockscope(lock->scope),
	     get_locktype(lock->type), get_timeout(lock->timeout),
	     lock->owner?lock->owner:_("(none)"), get_depth(lock->depth));
}

void execute_discover(const char *res)
{
    struct dav_lock_result *list, *cur, *next;
    char *real_remote;
    int ret;
    real_remote = resolve_path(path, res, false);
    out_start(_("Discovering locks on"), res);
    ret = dav_lock_discover(session, real_remote, &list);
    switch (ret) {
    case HTTP_OK:
	out_success();
	for (cur = list; cur != NULL; cur = next) {
	    struct dav_lock *lock, *nextlock;
	    for (lock = cur->lock; lock != NULL; lock = nextlock) {
		print_lock(lock);
		nextlock = lock->next;
		dav_lock_free(lock);
	    }
	    next = cur->next;
	    HTTP_FREE(cur->href);
	    free(cur);
	}

	break;
    default:
	out_result(ret);
	break;
    }
    free(real_remote);
}

void execute_steal(const char *res)
{
    struct dav_lock_result *list, *cur, *next;
    char *real_remote;
    int ret;
    real_remote = resolve_path(path, res, false);
    out_start(_("Stealing locks on"), res);
    ret = dav_lock_discover(session, real_remote, &list);
    switch (ret) {
    case HTTP_OK:
	out_success();
	for (cur = list; cur != NULL; cur = next) {
	    struct dav_lock *lock;
	    for (lock = cur->lock; lock != NULL; lock = lock->next) {
		printf("%s: <%s>\n", cur->href, cur->lock->token);
		dav_lock_add(lock_session, lock);
	    }
	    next = cur->next;
	    free(cur);
	}
	break;
    default:
	out_result(ret);
	break;
    }
}

static void showlocks_walker(struct dav_lock *lock, void *userdata)
{
    print_lock(lock);
}

void execute_showlocks(void)
{
    printf(_("List of owned locks:\n"));
    if (dav_lock_iterate(lock_session, showlocks_walker, NULL) == 0) {
	printf(_("No owned locks.\n"));
    }
}

void execute_lock(const char *res)
{
    char *real_remote;
    struct dav_lock *lock;
    lock = ne_calloc(sizeof *lock);
    out_start(_("Locking"), res);
    real_remote = true_path(res);
    lock->type = dav_locktype_write;
    lock->scope = dav_lockscope_exclusive;
    lock->owner = getowner();
    if (uri_has_trailing_slash(real_remote)) {
	lock->depth = DAV_DEPTH_INFINITE;
    } else {
	lock->depth = 0;
    }
    lock->uri = real_remote;
    if (out_handle(dav_lock(session,lock))) {
	/* Locktoken: <%s>\n", lock->token); */
	dav_lock_add(lock_session, lock);
    } else {
	free(real_remote);
	free(lock);
    }
}

void execute_unlock(const char *res)
{
    struct dav_lock *lock;
    char *real_remote;
    int isdummy = 0;
    out_start(_("Unlocking"), res);
    real_remote = true_path(res);
    lock = dav_lock_find(lock_session, real_remote);
    if (!lock) {
	lock = ne_calloc(sizeof *lock);
	lock->token = readline(_("Enter locktoken: "));
	if (!lock->token || strlen(lock->token) == 0) {
	    free(lock);
	    return;
	}
	lock->uri = real_remote;
	isdummy=1;
    }
    out_result(dav_unlock(session, lock));
    if (isdummy) {
	free(lock);
    } else {
	dav_lock_remove(lock_session, lock);
	dav_lock_free(lock);
    }
}

void execute_mkcol(const char *filename)
{
    char *remote;
    remote = resolve_path(path, filename, true);
    dispatch("Creating", filename, dav_mkcol, remote);
    free(remote);
}

void execute_propnames(const char *res)
{
#if 0
    char *remote, **names = NULL;
    remote = resolve_path(path, res, false);
    out_start("Fetching property names", res);
    if (out_handle(dav_prop_getnames(remote, &names))) { 
	char **name;
	for(name = names; *name != NULL; name++) {
	    printf("Property: %s\n", *name);
	    free(*name);
	}
	free(names);
    }
    free(remote);
#endif
}

static void remove_locks(const char *uri, int depth)
{
    struct dav_lock *lck;
    
    lck = dav_lock_find(lock_session, uri);
    if (lck != NULL) {
	dav_lock_remove(lock_session, lck);
    }

}

void execute_delete(const char *filename)
{
    char *remote = resolve_path(path, filename, false);
    out_start(_("Deleting"), filename);
    if (is_collection(remote)) {
	output(o_finish, 
_("is a collection resource.\n"
"The `rm' command cannot be used to delete a collection.\n"
"Use `rmcol %s' to delete this collection and ALL its contents.\n"), 
filename);
    } else {
	out_result(dav_delete(session, remote));
    }
    remove_locks(remote, 0);
    free(remote);
}

void execute_rmcol(const char *filename)
{
    char *remote;
    remote = resolve_path(path, filename, true);
    out_start(_("Deleting collection"), filename);
    if (!is_collection(remote)) {
	output(o_finish, 
	       _("is not a collection.\n"
		 "The `rmcol' command can only be used to delete collections.\n"
		 "Use `rm %s' to delete this resource.\n"), filename);
    } else {
	out_result(dav_delete(session, remote));
    }
    remove_locks(remote, DAV_DEPTH_INFINITE);
    free(remote);
}

/* Like resolve_path except more intelligent. */
static char *clever_path(const char *path, const char *src, 
			 const char *dest)
{
    char *ret;
    int src_is_coll, dest_is_coll;
    dest_is_coll = (dest[strlen(dest)-1] == '/');
    src_is_coll = (src[strlen(src)-1] == '/');
    if (strcmp(dest, ".") == 0) {
	ret = resolve_path(path, base_name(src), false);
    } else if (strcmp(dest, "..") == 0) {
	char *parent;
	parent = uri_parent(path);
	ret = resolve_path(parent, base_name(src), false);
	free(parent);
    } else if (!src_is_coll && dest_is_coll) {
	/* Moving a file to a collection... the destination should
	 * be the basename of file concated with the collection. */
	char *tmp;
	tmp = resolve_path(path, dest, true);
	CONCAT2(ret, tmp, base_name(src));
	free(tmp);
    } else {
	ret = resolve_path(path, dest, false);
    }
    return ret;
}

static void cat_callback(void *userdata, const char *buf, size_t len)
{
    FILE *f = userdata;
    if (fwrite(buf, len, 1, f) < 0) {
	perror("fwrite");
    }
}

static char *choose_pager(void)
{
    struct stat st;
    char *tmp;
    /* TODO: add an autoconf for less/more? */
    tmp = getenv("PAGER");
    if (tmp != NULL) {
	return tmp;
    } else if (stat("/usr/bin/less", &st) == 0) {
	return "/usr/bin/less";
    } else if (stat("/bin/less", &st) == 0) {
	return "/bin/less";
    } else {
	return "/bin/more";
    }
}

static FILE *spawn_pager(const char *pager)
{
    /* Create a pipe */
    return popen(pager, "w");
}

static void kill_pager(FILE *p)
{
    /* This blocks until the pager quits. */
    pclose(p);
}

void execute_less(const char *resource)
{
    char *real_res;
    char *pager;
    FILE *p;
    real_res = resolve_path(path, resource, false);
    pager = choose_pager();
    printf(_("Displaying `%s':\n"), real_res);
    p = spawn_pager(pager);
    if (p == NULL) {
	printf(_("Error! Could not spawn pager `%s':\n%s\n"), pager,
		 strerror(errno));
    } else {
	child_running = true;
	http_read_file(session, real_res, cat_callback, p);
	kill_pager(p); /* Blocks until the pager quits */
	child_running = false;
    }
}

void execute_cat(const char *resource)
{
    char *real_res = resolve_path(path, resource, false);
    printf(_("Displaying `%s':\n"), real_res);
    if (http_read_file(session, real_res, cat_callback, stdout) != HTTP_OK) {
	printf(_("Failed: %s\n"), http_get_error(session));
    }
}

/* Returns non-zero if given resource is not a collection resource.
 * This function MAY make a request to the server. */
int is_collection(const char *uri)
{
    struct resource *res = NULL;
    int ret = 0;
    /* TODO: just request resourcetype here. */
    ret = fetch_resource_list(session, uri, DAV_DEPTH_ZERO, 1, &res);
    if (ret == HTTP_OK) {
	if (res != NULL && uri_compare(uri, res->uri) == 0) {
	    if (res->type == resr_collection) {
		ret = 1;
	    } else {
		http_set_error(session, _("Not a collection resource."));
	    }
	} else {
	    /* FIXME: this error occurs when you do open /foo and get
	     * the response for /foo/ back */
	    http_set_error(session, _("Did not find a collection resource."));
	}
    } else {
	ret = 0;
    }
    free_resource_list(res);
    return ret;
}

void multi_copy(int argc, const char *argv[])
{
    do_copymove(argc, argv, _("copying"), _("copy"), simple_copy);
}

void multi_move(int argc, const char *argv[])
{
    do_copymove(argc, argv, _("moving"), _("move"), simple_move);
}

static void do_copymove(int argc, const char *argv[],
			const char *v1, const char *v2,
			void (*cb)(const char *, const char *))
{
    /* We are guaranteed that argc > 2... */
    char *dest;

    dest = resolve_path(path, argv[argc-1], true);
    if (is_collection(dest)) {
	int n;
	char *real_src, *real_dest;
	for(n = 0; n < argc-1; n++) {
	    real_src = resolve_path(path, argv[n], false);
	    real_dest = clever_path(path, argv[n], dest);
	    if (strcmp(real_src, real_dest) == 0) {
		printf("%s: %s and %s are the same resource.\n", v2,
			real_src, real_dest);
	    } else {
		(*cb)(real_src, real_dest);
	    }
	    free(real_src);
	    free(real_dest);
	}
    } else if (argc > 2) {
	printf(
	    _("When %s multiple resources, the last argument must be a collection.\n"), v1);
    } else {
	char *rsrc, *rdest;
	rsrc = resolve_path(path, argv[0], false);
	rdest = resolve_path(path, argv[1], false);
	/* Simple */
	(*cb) (rsrc, rdest);
	free(rsrc);
	free(rdest);
    }
    free(dest);
}

static void simple_move(const char *src, const char *dest) 
{
    output(o_start, _("Moving `%s' to `%s': "), src, dest);
    out_result(dav_move(session, 1, src, dest));
}

static void simple_copy(const char *src, const char *dest) 
{
    output(o_start, _("Copying `%s' to `%s': "), src, dest);
    out_result(dav_copy(session, 1, src, dest));
}

/* Returns absolute filename which is 'filename' relative to 
 * (absolute filename) 'path'. e.g.
 *    resolve_path("/dav/", "asda") == "/dav/asda"
 *    resolve_path("/dav/", "/asda") == "/asda"
 * Also removes '..' segments, e.g.
 *    resolve_path("/dav/foobar/", "../fish") == "/dav/fish"
 * If isdir is true, ensures the return value has a trailing slash.
 */
char *resolve_path(const char *path, const char *filename, int isdir)
{
    char *ret, *pnt;
    if (*filename == '/') {
	/* It's absolute */
	ret = ne_strdup(filename);
    } else if (strcmp(filename, ".") == 0) {
	ret = ne_strdup(path);
    } else {
	CONCAT2(ret, path, filename);
    }
    if (isdir && ret[strlen(ret)-1] != '/') {
	char *newret;
	CONCAT2(newret, ret, "/");
	free(ret);
	ret = newret;
    }
    /* Sort out '..', etc... */
    do {
	pnt = strstr(ret, "/../");
	if (pnt != NULL) {
	    char *last;
	    /* Find the *previous* path segment, to overwrite...
	     *    /foo/../
	     *    ^- last points here */
	    if (pnt > ret) {
		for(last = pnt-1; (last > ret) && (*last != '/'); last--);
	    } else {
		last = ret;
	    }
	    memmove(last, pnt + 3, strlen(pnt+2));
	} else {
	    pnt = strstr(ret, "/./");
	    if (pnt != NULL) {
		memmove(pnt, pnt+2, strlen(pnt+1));
	    }
	}
    } while (pnt != NULL);
    return ret;    
}

void execute_get(const char *remote, const char *local)
{
    char *filename, *real_remote;
    real_remote = resolve_path(path, remote, false);
    if (local == NULL) {
	struct stat st;
	/* Choose an appropriate local filename */
	if (stat(base_name(remote), &st) == 0) {
	    /* File already exists... don't overwrite */
	    printf(_("Enter local filename for `%s': "), real_remote);
	    filename = readline(NULL);
	    if (filename == NULL) {
		free(real_remote);
		printf(_("cancelled.\n"));
		return;
	    }
	} else {
	    filename = ne_strdup(base_name(remote));
	}
    } else {
	filename = ne_strdup(local);
    }
    {
	FILE *f = fopen(filename, "w");
	output(o_transfer, _("Downloading `%s' to %s:"), real_remote, filename);
	if (f == NULL) {
	    output(o_finish, _("failed:\n%s\n"), strerror(errno));
	} else {
	    out_result(http_get(session, real_remote, f));
	    /* FIXME */ fclose(f);
	}
    }
    free(real_remote);
    free(filename);
}

static int run_editor(const char *filename)
{
    char editcmd[BUFSIZ], *editor;
    struct stat before_st, after_st;
    editor = get_option(opt_editor);
    if (editor == NULL) {
	editor = getenv("EDITOR");
	if (editor == NULL) {
	    editor = "vi";
	}
    }
    snprintf(editcmd, BUFSIZ, "%s %s", editor, filename);
    if (stat(filename, &before_st)) {
	printf(_("Could not stat file: %s\n"), strerror(errno));
	return -1;
    }
    printf("Running editor: `%s'...\n", editcmd);
    system(editcmd);
    if (stat(filename, &after_st)) {
	printf(_("Error! Could not examine temporary file: %s\n"), 
		 strerror(errno));
	return -1;
    }
    if (before_st.st_mtime == after_st.st_mtime) {
	/* File not changed. */
	printf(_("No changes were made.\n"));
	return -1;
    } else {
	printf(_("Changes were made.\n"));
	return 0;
    }	
}

/* TODO: this is great big heap of steaming trout.  */
void execute_edit(const char *remote)
{
    char *real_remote;
    unsigned int can_lock; /* can we LOCK it? */
    struct dav_lock *lock;
    char *fname;
    FILE *f;
    
    real_remote = resolve_path(path, remote, false);

#if 0
    /* TODO: fetch 'supportedlock' and check whether we can really
     * lock it. */
    if (http_options(session, real_remote, &caps) != HTTP_OK) {
	printf("Could not determining locking capabilities for `%s':\n%s\n", 
		remote, http_get_error(session));
	return;
    }

    can_lock = http_webdav_locking;
#else
    can_lock = 1;
#endif

    /* Get a temporary filename.
     * FIXME: joe: my tmpname(3) man page says it is deprecated */
    fname = tmpnam(NULL);
    if (fname == NULL) {
	printf(_("Could not create temporary filename: %s\n"), 
	       strerror(errno));
	return;
    }
    f = fopen(fname, "w" FOPEN_BINARY_FLAGS);
    if (f == NULL) {
	printf(_("Could not open temporary file: %s\n"), strerror(errno));
	return;
    }
   
    if (can_lock) {
	/* FIXME: copy'n'friggin'paste */
	lock = ne_calloc(sizeof *lock);
	lock->type = dav_locktype_write;
	lock->scope = dav_lockscope_exclusive;
	lock->owner = getowner();
	lock->depth = 0;
	lock->uri = ne_strdup(real_remote);
	out_start("Locking", remote);
	if (out_handle(dav_lock(session, lock))) {
	    dav_lock_add(lock_session, lock);
	} else {
	    dav_lock_free(lock);
	    return;
	}
    } else {
	/* TODO: HEAD and get the Etag/modtime */
    }
    /* FIXME: copy'n'friggin'paste. */
    output(o_transfer, _("Downloading `%s' to %s"), real_remote, fname);
    if (out_handle(http_get(session, real_remote, f))) {
	if (fclose(f)) {
	    output(o_finish, _("Error writing to temporary file: %s\n"), 
		   strerror(errno));
	} else {
	    if (run_editor(fname)) {
		/* Edit failed */
	    } else {
		int upload_okay = 0;
		/* Edit succeeded. TODO: reopen the file? 
		 * Do editors unlink+create a new file? */
		f = fopen(fname, "r" FOPEN_BINARY_FLAGS);
		if (f == NULL) {
		    output(o_finish, 
			   _("Could not re-open temporary file: %s\n"),
			   strerror(errno));
		} else {
		    do {
			output(o_transfer, _("Uploading changes to `%s'"), 
			       real_remote);
			/* FIXME: conditional PUT using fetched Etag/modtime if
			 * !can_lock */
			if (out_handle(http_put(session, real_remote, f))) {
			    upload_okay = 1;
			} else {
			    /* TODO: offer to save locally instead */
			    printf(_("Try uploading again (y/n)? "));
			    if (!yesno()) {
				upload_okay = 1;
			    }
			}
		    } while (!upload_okay);
		    fclose(f);
		}
	    }
	}
    }
    if (unlink(fname)) {
	printf(_("Could not remove temporary file `%s': %s\n"),
		 fname, strerror(errno));
    }
    /* UNLOCK it again whether we succeed or failed in our mission */
    if (can_lock) {
	output(o_start, "Unlocking `%s':", remote);
	out_result(dav_unlock(session,lock));
	dav_lock_remove(lock_session, lock);
	dav_lock_free(lock);
    }

}

static void simple_put(const char *local, const char *remote)
{
    FILE *f = fopen(local, "r" FOPEN_BINARY_FLAGS);
    output(o_transfer, _("Uploading %s to `%s':"), local, remote);
    if (f == NULL) {
	output(o_finish, _("Could not open file: %s\n"), strerror(errno));
    } else {
	out_result(http_put(session, remote, f));
	(void) fclose(f);
    }
}

void execute_put(const char *local, const char *remote)
{
    char *real_remote;
    if (remote == NULL) {
	real_remote = resolve_path(path, base_name(local), false);
    } else {
	real_remote = resolve_path(path, remote, false);
    }
    simple_put(local, real_remote);
    free(real_remote);
}

/* this is getting too easy */
void multi_mput(int argc, const char *argv[])
{
    for(; argv[0] != NULL; argv++) {
	char *remote;
	remote = clever_path(path, argv[0], path);
	simple_put(argv[0], remote);
	free(remote);
    }    
}

void multi_mget(int argc, const char *argv[])
{
    for(; argv[0] != NULL; argv++) {
	execute_get(argv[0], NULL);
    }
}

void execute_chexec(const char *val, const char *remote)
{
    static const dav_propname execprop = 
    { "http://apache.org/dav/props/", "executable" };
    /* Use a single operation; set the executable property to... */
    dav_proppatch_operation ops[] = { { &execprop, dav_propset, NULL }, { NULL } };
    char *uri;
    http_server_capabilities caps = {0};
    int ret;

    /* True or false, depending... */
    if (strcmp(val, "+") == 0) {
	ops[0].value = "T";
    } else if (strcmp(val, "-") == 0) {
	ops[0].value = "F";
    } else {
	printf(_("Use:\n"
		 "   chexec + %s   to make the resource executable\n"
		 "or chexec - %s   to make the resource unexecutable\n"),
		 remote, remote);
	return;
    }
    
    uri = resolve_path(path, remote, false);

    out_start(_("Setting isexecutable"), remote);
    ret = http_options(session, uri, &caps);
    if (ret != HTTP_OK) {
	out_result(ret);
    } else if (!caps.dav_executable) {
	http_set_error(session, 
		       _("The server does not support the 'isexecutable' property."));
	out_result(HTTP_ERROR);
    } else {    
	out_result(dav_proppatch(session, uri, ops));
    }
    free(uri);

}

void execute_lpwd(void)
{
    char pwd[BUFSIZ];
    if (getcwd(pwd, BUFSIZ) == NULL) {
	perror("pwd");
    } else {
	printf(_("Local directory: %s\n"), pwd);
    }
}

/* Using a hack, we get zero-effort multiple-argument 'lls' */
void execute_lls(int argc, char * const argv[])
{
    int pid;
    pid = fork();
    switch (pid) {
    case 0: /* child...  */
	/* The implicit typecast munging might screw things up in
	 * weird and wonderful ways, here? */
	execvp("ls", &argv[-1]);
	printf(_("Error executing ls: %s\n"), strerror(errno));
	exit(-1);
	break;
    case -1: /* Error */
	printf(_("Error forking ls: %s\n"), strerror(errno));
	break;
    default: /* Parent */
	wait(NULL);
	break;
    }
    return;
}

void execute_lcd(const char *path)
{
    const char *real_path;
    if (path) {
	real_path = path;
    } else {
	real_path = getenv("HOME");
	if (!real_path) {
	    printf(_("Could not determine home directory from environment.\n"));
	    return;
	}
    }
    if (chdir(real_path)) {
	printf(_("Could not change local directory:\nchdir: %s\n"),
	       strerror(errno));
    }
}

void execute_pwd(void)
{
    printf(_("Current collection is `%s', on server `%s'\n"), path, 
	    server_hostname);
}

void execute_cd(const char *newpath)
{
    char *real_path;
    int is_swap = false;
    if (strcmp(newpath, "-") == 0) {
	is_swap = true;
	real_path = old_path;
    } else if (path == NULL) {
	/* The first time we are executing a CD. */
	if (1 || uri_has_trailing_slash(newpath)) {
	    real_path = ne_strdup(newpath);
	} else {
	    CONCAT2(real_path, newpath, "/");
	}
    } else {
	real_path = resolve_path(path, newpath, true);
    }
    if (set_path(real_path)) {
	/* Error */
	if (!is_swap) free(real_path);
    } else {
	/* Success */
	if (!is_swap && old_path) free(old_path);
	old_path = path;
	path = real_path;
    }
}

static void display_help_message(void)
{
    int n;
    printf("Commands: \n");
    
    for(n = 0; commands[n].id != cmd_unknown; n++) {
	if (commands[n].call && commands[n].short_help) {
	    printf(" %-26s %s\n", commands[n].call, _(commands[n].short_help));
	}
    }

    printf(_("Aliases: rm=delete, mkdir=mkcol, mv=move, cp=copy, "
	     "more=less, quit=exit=bye\n"));
}

void execute_help(const char *arg)
{
    if (!arg) {
	display_help_message();
    } else {
	const struct command *cmd = get_command(arg);
	if (cmd) {
	    printf(" `%s'   %s\n", cmd->call, cmd->short_help);
	    if (cmd->needs_connection) {
		printf(_("This command can only be used when connected to a server.\n"));
	    }
	} else {
	    printf(_("Command name not known: %s\n"), arg);
	}
    }
}

void execute_echo(int count, const char **args)
{
    const char **pnt;
    for(pnt = args; *pnt != NULL; pnt++) {
	printf("%s ", *pnt);
    }
    putchar('\n');
}

